"""Waveform timing data types for NI Python APIs."""

from nitypes.waveform._timing._sample_interval import SampleIntervalMode
from nitypes.waveform._timing._timing import Timing

__all__ = [
    "SampleIntervalMode",
    "Timing",
]
